@@ -122,6 +122,8 @@ G80SorModeSet(xf86OutputPtr output, DisplayModePtr mode,
         type |
         ((adjusted_mode->Flags & V_NHSYNC) ? 0x1000 : 0) |
         ((adjusted_mode->Flags & V_NVSYNC) ? 0x2000 : 0));
+
+    G80CrtcSetScale(output->crtc, adjusted_mode, pPriv->scale);
 }
 
 static xf86OutputStatus
@@ -151,49 +153,90 @@ G80SorDestroy(xf86OutputPtr output)
 
     G80OutputDestroy(output);
 
-    if(pPriv->nativeMode) {
-        if(pPriv->nativeMode->name)
-            xfree(pPriv->nativeMode->name);
-        xfree(pPriv->nativeMode);
-    }
+    xf86DeleteMode(&pPriv->nativeMode, pPriv->nativeMode);
 
     xfree(output->driver_private);
     output->driver_private = NULL;
 }
 
-/******************** LVDS ********************/
+static void G80SorSetModeBackend(DisplayModePtr dst, const DisplayModePtr src)
+{
+    // Stash the backend mode timings from src into dst
+    dst->Clock           = src->Clock;
+    dst->Flags           = src->Flags;
+    dst->CrtcHDisplay    = src->CrtcHDisplay;
+    dst->CrtcHBlankStart = src->CrtcHBlankStart;
+    dst->CrtcHSyncStart  = src->CrtcHSyncStart;
+    dst->CrtcHSyncEnd    = src->CrtcHSyncEnd;
+    dst->CrtcHBlankEnd   = src->CrtcHBlankEnd;
+    dst->CrtcHTotal      = src->CrtcHTotal;
+    dst->CrtcHSkew       = src->CrtcHSkew;
+    dst->CrtcVDisplay    = src->CrtcVDisplay;
+    dst->CrtcVBlankStart = src->CrtcVBlankStart;
+    dst->CrtcVSyncStart  = src->CrtcVSyncStart;
+    dst->CrtcVSyncEnd    = src->CrtcVSyncEnd;
+    dst->CrtcVBlankEnd   = src->CrtcVBlankEnd;
+    dst->CrtcVTotal      = src->CrtcVTotal;
+    dst->CrtcHAdjusted   = src->CrtcHAdjusted;
+    dst->CrtcVAdjusted   = src->CrtcVAdjusted;
+}
+
 static Bool
-G80SorModeFixupScale(xf86OutputPtr output, DisplayModePtr mode,
-                     DisplayModePtr adjusted_mode)
+G80SorModeFixup(xf86OutputPtr output, DisplayModePtr mode,
+                DisplayModePtr adjusted_mode)
 {
     G80OutputPrivPtr pPriv = output->driver_private;
     DisplayModePtr native = pPriv->nativeMode;
 
-    // Stash the saved mode timings in adjusted_mode
-    adjusted_mode->Clock = native->Clock;
-    adjusted_mode->Flags = native->Flags;
-    adjusted_mode->CrtcHDisplay = native->CrtcHDisplay;
-    adjusted_mode->CrtcHBlankStart = native->CrtcHBlankStart;
-    adjusted_mode->CrtcHSyncStart = native->CrtcHSyncStart;
-    adjusted_mode->CrtcHSyncEnd = native->CrtcHSyncEnd;
-    adjusted_mode->CrtcHBlankEnd = native->CrtcHBlankEnd;
-    adjusted_mode->CrtcHTotal = native->CrtcHTotal;
-    adjusted_mode->CrtcHSkew = native->CrtcHSkew;
-    adjusted_mode->CrtcVDisplay = native->CrtcVDisplay;
-    adjusted_mode->CrtcVBlankStart = native->CrtcVBlankStart;
-    adjusted_mode->CrtcVSyncStart = native->CrtcVSyncStart;
-    adjusted_mode->CrtcVSyncEnd = native->CrtcVSyncEnd;
-    adjusted_mode->CrtcVBlankEnd = native->CrtcVBlankEnd;
-    adjusted_mode->CrtcVTotal = native->CrtcVTotal;
-    adjusted_mode->CrtcHAdjusted = native->CrtcHAdjusted;
-    adjusted_mode->CrtcVAdjusted = native->CrtcVAdjusted;
-
-    // This mode is already "fixed"
-    G80CrtcSkipModeFixup(output->crtc);
+    if(native && pPriv->scale != G80_SCALE_OFF) {
+        G80SorSetModeBackend(adjusted_mode, native);
+        // This mode is already "fixed"
+        G80CrtcSkipModeFixup(output->crtc);
+    }
 
     return TRUE;
 }
 
+static Bool
+G80SorTMDSModeFixup(xf86OutputPtr output, DisplayModePtr mode,
+                    DisplayModePtr adjusted_mode)
+{
+    int scrnIndex = output->scrn->scrnIndex;
+    G80OutputPrivPtr pPriv = output->driver_private;
+    DisplayModePtr modes = output->probed_modes;
+
+    xf86DeleteMode(&pPriv->nativeMode, pPriv->nativeMode);
+
+    if(modes) {
+        // Find the preferred mode and use that as the "native" mode.
+        // If no preferred mode is available, use the first one.
+        DisplayModePtr mode;
+
+        // Find the preferred mode.
+        for(mode = modes; mode; mode = mode->next) {
+            if(mode->type & M_T_PREFERRED) {
+                xf86DrvMsgVerb(scrnIndex, X_INFO, 5,
+                               "%s: preferred mode is %s\n",
+                               output->name, mode->name);
+                break;
+            }
+        }
+
+        // XXX: May not want to allow scaling if no preferred mode is found.
+        if(!mode) {
+            mode = modes;
+            xf86DrvMsgVerb(scrnIndex, X_INFO, 5,
+                    "%s: no preferred mode found, using %s\n",
+                    output->name, mode->name);
+        }
+
+        pPriv->nativeMode = xf86DuplicateMode(mode);
+        G80CrtcDoModeFixup(pPriv->nativeMode, mode);
+    }
+
+    return G80SorModeFixup(output, mode, adjusted_mode);
+}
+
 static DisplayModePtr
 G80SorGetLVDSModes(xf86OutputPtr output)
 {
@@ -211,6 +254,7 @@ struct property {
 
 static struct {
     struct property dither;
+    struct property scale;
 } properties;
 
 static void
@@ -219,7 +263,9 @@ G80SorCreateResources(xf86OutputPtr output)
     ScrnInfoPtr pScrn = output->scrn;
     G80Ptr pNv = G80PTR(pScrn);
     int data, err;
+    const char *s;
 
+    /******** dithering ********/
     properties.dither.atom = MAKE_ATOM("dither");
     properties.dither.range[0] = 0;
     properties.dither.range[1] = 1;
@@ -240,11 +286,33 @@ G80SorCreateResources(xf86OutputPtr output)
         xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                    "Failed to set dithering property for %s: error %d\n",
                    output->name, err);
+
+    /******** scaling ********/
+    properties.scale.atom = MAKE_ATOM("scale");
+    err = RRConfigureOutputProperty(output->randr_output,
+                                    properties.scale.atom, FALSE, FALSE,
+                                    FALSE, 0, NULL);
+    if(err)
+        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+                   "Failed to configure scaling property for %s: error %d\n",
+                   output->name, err);
+
+    // Set the default value
+    s = "aspect";
+    err = RRChangeOutputProperty(output->randr_output, properties.scale.atom,
+                                 XA_STRING, 8, PropModeReplace, strlen(s),
+                                 (pointer)s, FALSE, FALSE);
+    if(err)
+        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+                   "Failed to set scaling property for %s: error %d\n",
+                   output->name, err);
 }
 
 static Bool
 G80SorSetProperty(xf86OutputPtr output, Atom prop, RRPropertyValuePtr val)
 {
+    G80OutputPrivPtr pPriv = output->driver_private;
+
     if(prop == properties.dither.atom) {
         INT32 i;
 
@@ -256,9 +324,69 @@ G80SorSetProperty(xf86OutputPtr output, Atom prop, RRPropertyValuePtr val)
             return FALSE;
 
         G80CrtcSetDither(output->crtc, i, TRUE);
+        return TRUE;
+    } else if(prop == properties.scale.atom) {
+        const char *s;
+        enum G80ScaleMode oldScale, scale;
+        int i;
+        const struct {
+            const char *name;
+            enum G80ScaleMode scale;
+        } modes[] = {
+            { "off",    G80_SCALE_OFF },
+            { "aspect", G80_SCALE_ASPECT },
+            { "fill",   G80_SCALE_FILL },
+            { "center", G80_SCALE_CENTER },
+            { NULL,     0 },
+        };
+
+        if(val->type != XA_STRING || val->format != 8)
+            return FALSE;
+        s = (char*)val->data;
+
+        for(i = 0; modes[i].name; i++) {
+            const char *name = modes[i].name;
+            const int len = strlen(name);
+
+            if(val->size == len && !strncmp(name, s, len)) {
+                scale = modes[i].scale;
+                break;
+            }
+        }
+        if(!modes[i].name)
+            return FALSE;
+        if(scale == G80_SCALE_OFF && pPriv->panelType == LVDS)
+            // LVDS requires scaling
+            return FALSE;
+
+        oldScale = pPriv->scale;
+        pPriv->scale = scale;
+        if(output->crtc) {
+            xf86CrtcPtr crtc = output->crtc;
+
+            if(!xf86CrtcSetMode(crtc, &crtc->desiredMode, crtc->desiredRotation,
+                                crtc->desiredX, crtc->desiredY)) {
+                xf86DrvMsg(crtc->scrn->scrnIndex, X_ERROR,
+                           "Failed to set scaling to %s for output %s\n",
+                           modes[i].name, output->name);
+
+                // Restore old scale and try again.
+                pPriv->scale = oldScale;
+                if(!xf86CrtcSetMode(crtc, &crtc->desiredMode,
+                                    crtc->desiredRotation, crtc->desiredX,
+                                    crtc->desiredY)) {
+                    xf86DrvMsg(crtc->scrn->scrnIndex, X_ERROR,
+                               "Failed to restore old scaling for output %s\n",
+                               output->name);
+                }
+
+                return FALSE;
+            }
+        }
+        return TRUE;
     }
 
-    return TRUE;
+    return FALSE;
 }
 #endif // RANDR_12_INTERFACE
 
@@ -267,7 +395,7 @@ static const xf86OutputFuncsRec G80SorTMDSOutputFuncs = {
     .save = NULL,
     .restore = NULL,
     .mode_valid = G80TMDSModeValid,
-    .mode_fixup = G80OutputModeFixup,
+    .mode_fixup = G80SorTMDSModeFixup,
     .prepare = G80OutputPrepare,
     .commit = G80OutputCommit,
     .mode_set = G80SorModeSet,
@@ -285,7 +413,7 @@ static const xf86OutputFuncsRec G80SorLVDSOutputFuncs = {
     .save = NULL,
     .restore = NULL,
     .mode_valid = G80LVDSModeValid,
-    .mode_fixup = G80SorModeFixupScale,
+    .mode_fixup = G80SorModeFixup,
     .prepare = G80OutputPrepare,
     .commit = G80OutputCommit,
     .mode_set = G80SorModeSet,
